#import "WiiMoteControlPatch.h"
#import "WiiMoteControlUI.h"

@implementation WiiMoteControlPatch : QCPatch

+ (int)executionMode
{
	return 2;
}

+ (BOOL)allowsSubpatches
{
	return NO;
}

+ (int)timeMode
{
	return 1;
}

+ (Class)inspectorClassWithIdentifier:(id)fp8
{
	return [WiiMoteControlUI class];
}

- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];

	// this is probably pedantic, but better safe than sorry :)
	[remoteLock lock];
		remote = nil;
	[remoteLock unlock];

	return z;
}

- (id)setup:(id)fp8
{
	BluetoothHCIPowerState powerState;
	remoteLock = [[NSLock alloc] init];

	//NSLog(@"WiiMote Setup\n");
	
	// franz noticed that if bluetooth is off, things freak out.  Pretty annoying.
	// so, to work around the bug in the DarWiinRemote framework, we need to see if bluetooth power
	// is off, and skip init if it is.
	
	// start wiimote discovery service
	if(IOBluetoothLocalDeviceAvailable())
	{
		IOBluetoothLocalDeviceGetPowerState(&powerState);
		if(powerState == kBluetoothHCIPowerStateON)
			discover = [[WiiRemoteDiscovery alloc] init];
		else
			discover = nil;
		if(discover != nil)
		{
			//NSLog(@"WiiMote configuring BT discovery...\n");
			[discover setDelegate:self];
			[discover start];
			// Notify us when expansion port changes happen
			[[NSNotificationCenter defaultCenter] addObserver:self
				selector:@selector(expansionPortChanged:)
				name:@"WiiRemoteExpansionPortChangedNotification"
				object:nil];
		}
	}

	//NSLog(@"WiiMote Setup Complete\n");

	return fp8;
}

- (void)cleanup:(id)fp8
{
	//NSLog(@"WiiMote Cleanup\n");
	[remoteLock lock];
		if(remote)
		{
			[remote closeConnection];
			[remote release];
			remote = nil;
		}
	[remoteLock unlock];
	
	if(discover)
		[discover release];
	[remoteLock release];
}

- (void)enable:(id)fp8
{
	//NSLog(@"WiiMote Enabled\n");
}

- (void)disable:(id)fp8
{
	//NSLog(@"WiiMote Disabled\n");
}


- (BOOL)execute:(id)fp8 time:(double)fp12 arguments:(id)fp20
{
	//NSLog(@"WiiMote: execute\n");
	// Error handing here serves no purpose!  So, why crash when we can do nothing?!  :)
	/*if([remoteLock tryLock] == YES)
	{
		if( remote != nil && [remote available] == NO)
		{
			if(remote != nil)
			{
				// at this point we crash quartz... need to debug (it's hard to get here though.. :/)
				// it looks like remote is stale?  it's non-nil, but invalid too.
				NSLog(@"Terminating connection (remote=0x%08x):  This will get messy\n",remote);
				//[remote closeConnection];
				[remote release];
				NSLog(@"   * Released remote\n");
				[outputWiiMoteConnected setBooleanValue:NO];
				NSLog(@"   * Cleared Connected Status\n");
				[discover start];
				NSLog(@"   * Discovery restarted\n");
			}
			// at this point, we're trying to reconnect.  No point handling buttons and stuff :/
			[remoteLock unlock];
			return YES;
		}
		[remoteLock unlock];
	}*/
	
	// This whole hack in execute is due to a stupid first-connect bug.
	// the current framework auto-disconnects a freshly connect remote on its first go,
	// so we we have this check here after 1 second to see if it got auto disconnected.  if
	// it did, we disconnect and restart the discovery process.
	[remoteLock lock];
	//NSLog(@"time: %d\nold : %d\n",time(0),discoverTime);
	if(time(0) > discoverTime && remote != nil)
	{
		if([remote available] == YES)
		{
			//NSLog(@"WiiMote became available\n");
			discoverTime = time(0) + 10;	// only log every 10 seconds to not flood logs...
		}
		else
		{
			NSLog(@"WiiMote seems to have disappeared...\n");
			[outputWiiMoteConnected setBooleanValue:NO];
			// Must closeConnection, or the remote thinks it's connected until it times out.
			[remote closeConnection];
			[remote release];
			remote = nil;
			if(discover)
			{
				[discover setDelegate:self];
				[discover start];
			}
		}
	}
	[remoteLock unlock];
	
	// once the above is fixed, the following will be true:
	// since everything is asyncronous, we don't actually need to do anything! :)  yay 60fps :)
	return YES;
}

// delegate methods
- (void) WiiRemoteDiscovered:(WiiRemote*)wiimote
{
	if(remote)
		{NSLog(@"Duplicate Discovery\n");return;}	// got here again?  ignore it (hasn't happened yet)
	//NSLog(@"Discovered WiiMote\n");
	[remoteLock lock];
		remote = wiimote;
		[remote setDelegate:self];
		
		// There are gobs of NSLogs here because some events start happening before discovery is complete.
		// This isn't a problem normally (remoteLock saves us), except for the Disconnect method below.
		// It happens on fresh connects, and if it tries to acquire the lock, it blocks, causing a deadlock
		// that hangs quartz.  meh.  That's where there's the stupid "remote != nil" check a few lines down...
		// We can walk through the logs to see how far along things get incase disaster (deadlock) strikes.
		
		discoverTime = time(0);
		//[remote setDelegate:self];
		/*if([remote available])
			NSLog(@"WiiMote reports availablity\n");*/
		// We'll set X00X LED's so we know it's quartzified
		[remote setLEDEnabled1:YES enabled2:NO enabled3:NO enabled4:YES];
		//NSLog(@"WiiMote LEDs configured\n");
		//usleep(10000);
		[remote setMotionSensorEnabled:YES];
		//NSLog(@"WiiMote Motion Sensor Enabled\n");
		//usleep(10000);
		//[remote setIRSensorEnabled:YES];
		//NSLog(@"WiiMote InfraRed Sensor Enabled\n");
		//usleep(10000);		
		/*if([remote isExpansionPortAttached])
			[remote setExpansionPortEnabled:YES];*/
		//NSLog(@"WiiMote Expansion Port Inspected and Configured\n");
		// sometimes on initial connections disconnect gets called before we finish
		// initializing.  In that case, we don't want to disable the discovery service.
		if(remote != nil)
			[discover stop];
		
		/*if([remote available])
			NSLog(@"WiiMote reports availablity\n");*/
	[remoteLock unlock];
	[outputWiiMoteConnected setBooleanValue:YES];
	
	//NSLog(@"Discovery Complete\n");
}

// from the autonotifyer.  When someone plugs in or unplugs something.
// Seems to get called twice per event most of the time, but once on bootup sometimes?
-(void) expansionPortChanged:(NSNotification *)notify
{
	//NSLog(@"expansionPortChanged\n");
	return;
	
	/*[remoteLock lock];
	if([remote isExpansionPortAttached])
	{
		[remote setExpansionPortEnabled:YES];
		[remoteLock unlock];
	}
	else
	{
		[remote setExpansionPortEnabled:NO];
		[remoteLock unlock];
		// clear nunchuk values on unplug
		[outputNunchukXAccel setDoubleValue:0];
		[outputNunchukYAccel setDoubleValue:0];
		[outputNunchukZAccel setDoubleValue:0];
		[outputXTilt setDoubleValue:0.5];
		[outputYTilt setDoubleValue:0.5];
		[outputCButton setBooleanValue:NO];
		[outputZButton setBooleanValue:NO];
	}*/
}

- (void) WiiRemoteDiscoveryError:(int)code
{
	NSLog(@"WiiMote Discovery Error (0x%08x)\n",code);
	
	// restarting the discovery service and acquiring the remoteLock here
	// seem to cause a deadlock.  without the lock (i.e. just restarting
	// the discovery service) we never get reconnected.
	// 
	// oddly enough, by not doing anything we have the best success rate.
	// *shrugs*
	
	/*[discover stop];
	
	[remoteLock lock];
		if(remote != nil)
		{
			//[remote closeConnection];
			[remote release];
			remote = nil;
		}
	[remoteLock unlock];
	
	[discover start];*/
}

// If this only happened when a remote died, things would be quite nice.
// However, as noted elsewhere it gets called shortly after connection on freshly-booted
// remotes, often before discovery completes.  read: race-condition/deadlock magnet!
- (void) wiiRemoteDisconnected:(IOBluetoothDevice*)device
{
	NSLog(@"WiiMote Disconnected\n");
	[outputWiiMoteConnected setBooleanValue:NO];
	//[remoteLock lock];
		[remote release];
		remote = nil;
	//[remoteLock unlock];
	NSLog(@"ReDiscovering WiiMotes\n");
	if(discover)
		[discover start];
}

// more methods
- (void) accelerationChanged:(WiiAccelerationSensorType)type accX:(unsigned char)accX accY:(unsigned char)accY accZ:(unsigned char)accZ wiiRemote:(WiiRemote*)wiiRemote
{
	if(type == WiiRemoteAccelerationSensor)
	{
		[remoteLock lock];
			WiiAccCalibData calib = [remote accCalibData:WiiRemoteAccelerationSensor];
		[remoteLock unlock];
		//NSLog here would be fatal.  this method gets called each time an update occurs, which is several times per second
		[outputXAccel setDoubleValue:(((double) accX)-127.0)/127.0];
		[outputYAccel setDoubleValue:(((double) accY)-127.0)/127.0];
		[outputZAccel setDoubleValue:(((double) accZ)-127.0)/127.0];

		// calculate tilt stuff
		// for now, we rely on built-in calibration stuff
		// we also smooth the values a tiny bit so its less spastic
		pitch = (0.7*pitch) + (0.3*atan( (double)(accY - calib.accY_zero)/(calib.accY_1g - calib.accY_zero) ) * 180.0 / 3.141592 * 2.0);
		roll =  (0.7*roll) +  (0.3*atan( (double)(accX - calib.accX_zero)/(calib.accX_1g - calib.accX_zero) ) * 180.0 / 3.141592 * 2.0);
		[outputPitch setDoubleValue: pitch];
		[outputRoll setDoubleValue: roll];
	}
	else	// must be a nunchuk update
	{
		/*[remoteLock lock];
			// This looks completely insane... but it seems to work?
			// the framework omits calibration data from the nunchuk for now, so we have to use
			// the fudged-but-closer remote calibration data (nunchuk's calib data is all zeros, so tilt is always 180.  boring)
			//WiiAccCalibData calib = [remote accCalibData:WiiNunchukAccelerationSensor];
			WiiAccCalibData calib = [remote accCalibData:WiiRemoteAccelerationSensor];
		[remoteLock unlock];
		
		[outputNunchukXAccel setDoubleValue:(((double)accX)-127.0)/127.0];
		[outputNunchukYAccel setDoubleValue:(((double)accY)-127.0)/127.0];
		[outputNunchukZAccel setDoubleValue:(((double)accZ)-127.0)/127.0];
		
		// calculate tilt stuff
		// for now, we rely on built-in calibration stuff
		// we also smooth the values a tiny bit so its less spastic
		// calibration data for nunchuk might be inaccurate.
		nunchukPitch = (0.7*nunchukPitch) + (0.3*atan( (double)(accY - calib.accY_zero)/(calib.accY_1g - calib.accY_zero) ) * 180.0 / 3.141592 * 2.0);
		nunchukRoll =  (0.7*nunchukRoll) +  (0.3*atan( (double)(accX - calib.accX_zero)/(calib.accX_1g - calib.accX_zero) ) * 180.0 / 3.141592 * 2.0);
		[outputNunchukPitch setDoubleValue: nunchukPitch];
		[outputNunchukRoll setDoubleValue: nunchukRoll];*/
	}

	return;
}

// pressure sensitive buttons!  How should we export this data?
- (void) analogButtonChanged:(WiiButtonType)type amount:(unsigned char)pressure wiiRemote:(WiiRemote*)wiiRemote{

}

- (void) joyStickChanged:(WiiJoyStickType)type tiltX:(unsigned char)tiltX tiltY:(unsigned char)tiltY wiiRemote:(WiiRemote*)wiiRemote
{
/*	if(type == WiiNunchukJoyStick)
	{
		[outputXTilt setDoubleValue:(double)tiltX];
		[outputYTilt setDoubleValue:(double)tiltY];
	}*/
}

- (void) buttonChanged:(WiiButtonType)type isPressed:(BOOL)isPressed wiiRemote:(WiiRemote*)wiiRemote
{
	//NSLog(@"buttonChanged\n");
	if(type == WiiRemoteAButton)
		[outputAButton setBooleanValue:isPressed];
	else if(type == WiiRemoteBButton)
		[outputBButton setBooleanValue:isPressed];
	else if(type == WiiRemoteUpButton)
		[outputUpButton setBooleanValue:isPressed];
	else if(type == WiiRemoteDownButton)
		[outputDownButton setBooleanValue:isPressed];
	else if(type == WiiRemoteLeftButton)
		[outputLeftButton setBooleanValue:isPressed];
	else if(type == WiiRemoteRightButton)
		[outputRightButton setBooleanValue:isPressed];
	else if(type == WiiRemoteMinusButton)
		[outputMinusButton setBooleanValue:isPressed];
	else if(type == WiiRemotePlusButton)
		[outputPlusButton setBooleanValue:isPressed];
	else if(type == WiiRemoteHomeButton)
		[outputHomeButton setBooleanValue:isPressed];
	else if(type == WiiRemoteOneButton)
		[output1Button setBooleanValue:isPressed];
	else if(type == WiiRemoteTwoButton)
		[output2Button setBooleanValue:isPressed];
	/*else if(type == WiiNunchukCButton)
		[outputCButton setBooleanValue:isPressed];
	else if(type == WiiNunchukZButton)
		[outputZButton setBooleanValue:isPressed];*/
}

// This may or may not get initialized.  The framework is pretty flakey about it for now.
- (void) irPointMovedX:(float)x Y:(float)y wiiRemote:(WiiRemote*)wiiRemote
{
	//[outputIRX setDoubleValue:x];
	//[outputIRY setDoubleValue:y];
}

// Full set of delegate methods from WiiRemote
/*
- (void) irPointMovedX:(float)px Y:(float)py wiiRemote:(WiiRemote*)wiiRemote;
- (void) rawIRData: (IRData[4])irData wiiRemote:(WiiRemote*)wiiRemote;
- (void) buttonChanged:(WiiButtonType)type isPressed:(BOOL)isPressed wiiRemote:(WiiRemote*)wiiRemote;
- (void) accelerationChanged:(WiiAccelerationSensorType)type accX:(unsigned char)accX accY:(unsigned char)accY accZ:(unsigned char)accZ wiiRemote:(WiiRemote*)wiiRemote;
- (void) joyStickChanged:(WiiJoyStickType)type tiltX:(unsigned char)tiltX tiltY:(unsigned char)tiltY wiiRemote:(WiiRemote*)wiiRemote;
- (void) analogButtonChanged:(WiiButtonType)type amount:(unsigned)press wiiRemote:(WiiRemote*)wiiRemote;
- (void) wiiRemoteDisconnected:(IOBluetoothDevice*)device;
*/

@end
